﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;
using OpenCvSharp;
using YSoft.Rqa.VisualProwess.AutoCalibration.Db.Entities.Base;
using Point = YSoft.Rqa.VisualProwess.AutoCalibration.Db.Entities.Base.Point;
using Size = OpenCvSharp.Size;

namespace YSoft.Rqa.VisualProwess.AutoCalibration.Db.Entities
{
    public class Camera: IdBase
    {
        public virtual List<MatrixElement> K { get; set; }
        public DistortionVector Distortion { get; set; }
        
        public static async Task<Camera> GetById(int id, CalibrationContext context)
        {
            var cam = await context.Cameras.FindAsync(id);
            if (cam == null)
                throw new ArgumentOutOfRangeException($"There is no camera with Id = {id}.");
            return cam;
        }

        public double[,] GetIntrinsicMatrix()
        {
            var k = new double[3, 3];
            K.ForEach(e => k[e.X, e.Y] = e.Value);
            return k;
        }

        public double[] GetDistortionCoefficients()
        {
            return new double[]
            {
                Distortion.k1,
                Distortion.k2,
                Distortion.p1,
                Distortion.p2,
                Distortion.k3,
                Distortion.k4,
                Distortion.k5,
                Distortion.k6,
                Distortion.s1,
                Distortion.s2,
                Distortion.s3,
                Distortion.s4
            };
        }

        public void ChangeK(double cx, double cy, double fx, double fy)
        {
            var Cx = K.Find(k => k.X == 0 && k.Y == 2);
            Cx.Value = Cx.Value + cx;
            var Cy = K.Find(k => k.X == 1 && k.Y == 2);
            Cy.Value = Cy.Value + cy;
            var Fx = K.Find(k => k.X == 0 && k.Y == 0);
            Fx.Value = Fx.Value + fx;
            var Fy = K.Find(k => k.X == 1 && k.Y == 1);
            Fy.Value = Fy.Value + fy;

        }

        public static double[,] GetIntrinsicParameters(List<Mat> inputImages, List<Point3f> chessboardModel, Size patternSize, out double[] distortion, bool manualControl)
        {
            var modelPoints = new List<List<Point3f>>();
            var imgPoints = new List<List<Point2f>>();
            var i = 0;
            foreach (var img in inputImages)
            {
                var isFound = Cv2.FindChessboardCorners(img, patternSize, out var corners);
                Console.Write(i + ": " + isFound);
                i++;
                if (!isFound)
                {
                    Console.WriteLine("");
                    continue;
                }
                var subCorners = Cv2.CornerSubPix(img.CvtColor(ColorConversionCodes.RGB2GRAY, 1), corners,
                        new Size(11, 11), new Size(-1, -1),
                        new TermCriteria(CriteriaType.Eps | CriteriaType.MaxIter, 500, 1e-6))
                    .ToList();

                if (manualControl)
                {
                    var img2Draw = img.Clone();
                    foreach (OpenCvSharp.Point point in subCorners)
                    {
                        Cv2.Circle(img2Draw, point, 3, new Scalar(10, 200, 50));
                        Cv2.ImShow("subCorners", img2Draw);
                        Cv2.WaitKey(0);
                    }

                    Cv2.ImWrite(@"C:\Users\adamek\Code\ysoft\improc\src\AutoCalibration\AutoCalibration\shots\manControl\" + Guid.NewGuid() + ".bmp", img2Draw);

                    var key = Cv2.WaitKey(0);

                    if (key == 13)
                    {
                        imgPoints.Add(subCorners);
                        modelPoints.Add(chessboardModel);
                        Console.WriteLine(" : Confirmed");
                    }
                    else Console.WriteLine(" : Not Confirmed");
                }
                else
                {
                    imgPoints.Add(subCorners);
                    modelPoints.Add(chessboardModel);
                    Console.WriteLine("");
                }
            }
            Console.WriteLine("_____________");
            Console.WriteLine("number of Confirmed: " + imgPoints.Count);
            var intrinsic = new double[3, 3];
            distortion = new double[12];
            var calFlag = CalibrationFlags.ThinPrismModel | CalibrationFlags.RationalModel;
            Cv2.CalibrateCamera(modelPoints, imgPoints, inputImages[0].Size(), intrinsic,
                distortion, out var rvec, out var tvec, calFlag);
            return intrinsic;
        }

        public Mat UndistortImage(Mat imageIn)
        {
            if (imageIn == null) throw new ArgumentNullException("Input image cannot be null");
            var imageOut = new Mat(imageIn.Size(), imageIn.Type());
            var newCamMat = Cv2.GetOptimalNewCameraMatrix(GetIntrinsicMatrix(), GetDistortionCoefficients(), imageIn.Size(), 0,
                imageIn.Size(), out _);
            var intr = new Mat(3, 3, MatType.CV_64F, GetIntrinsicMatrix());
            var dist = new Mat(GetDistortionCoefficients().Length, 1, MatType.CV_64F, GetDistortionCoefficients());
            var newMat = new Mat(3, 3, MatType.CV_64F, newCamMat);
            Cv2.Undistort(imageIn, imageOut, intr, dist, newMat);
            return imageOut;
        }

        public static void AddToDb(double[,] intrinsic, double[] distortion, CalibrationContext context)
        {
            var dist = new double[12];
            if (distortion.Length < 12)
            {
                var x = new double[12 - distortion.Length];
                dist = distortion.Concat(x).ToArray();
            }
            else
            {
                dist = distortion;
            }

            context.Cameras.Add(new Db.Entities.Camera
            {
                Distortion = new DistortionVector(dist),
                K = new List<MatrixElement>
                    {
                        new MatrixElement {X = 0, Y = 0, Value = intrinsic[0, 0]},
                        new MatrixElement {X = 0, Y = 1, Value = intrinsic[0, 1]},
                        new MatrixElement {X = 0, Y = 2, Value = intrinsic[0, 2]},
                        new MatrixElement {X = 1, Y = 0, Value = intrinsic[1, 0]},
                        new MatrixElement {X = 1, Y = 1, Value = intrinsic[1, 1]},
                        new MatrixElement {X = 1, Y = 2, Value = intrinsic[1, 2]},
                        new MatrixElement {X = 2, Y = 0, Value = intrinsic[2, 0]},
                        new MatrixElement {X = 2, Y = 1, Value = intrinsic[2, 1]},
                        new MatrixElement {X = 2, Y = 2, Value = intrinsic[2, 2]}
                    }
            });
            context.SaveChanges();
        }
    }
}