namespace FSLibrary
open MathNet.Numerics.LinearAlgebra

module GaussNewton = 

    let rad2deg a = a/(System.Math.PI / 180.0)
    let deg2rad a = a*(System.Math.PI / 180.0)
    
    let get_Rx a = //let a = deg2rad a
                   matrix [ [1.0; 0.0; 0.0];
                            [0.0; cos a; -sin a];
                            [0.0; sin a; cos a] ]
    let get_Ry b = //let b = deg2rad b
                   matrix [ [cos b; 0.0; sin b];
                            [0.0; 1.0; 0.0];
                            [-sin b; 0.0; cos b] ]
    let get_Rz c = //let c = deg2rad c
                   matrix [ [cos c; -sin c; 0.0];
                            [sin c; cos c; 0.0];
                            [0.0; 0.0; 1.0] ]

    let get_dRx a = //let a = deg2rad a
                    matrix [ [0.0; 0.0; 0.0];
                             [0.0; -sin a; -cos a];
                             [0.0; cos a; -sin a] ]
    let get_dRy b = //let b = deg2rad b
                    matrix [ [-sin b; 0.0; cos b];
                             [0.0; 0.0; 0.0];
                             [-cos b; 0.0; -sin b] ]
    let get_dRz c = //let c = deg2rad c
                    matrix [ [-sin c; -cos c; 0.0];
                             [cos c; -sin c; 0.0];
                             [0.0; 0.0; 0.0] ]

    let get_R(angles:double Matrix) = 
        get_Rz angles.[2,0] * get_Ry angles.[1,0] * get_Rx angles.[0,0]

    let get_P(K: double Matrix) (prms:double Matrix) = 
        let R = get_R prms.[3..5,0..0]
        let Rt = DenseMatrix.zero<double> 3 4
        Rt.[0..2,0..2] <- R
        Rt.[0..2,3..3] <- matrix [[prms.[0,0]];[prms.[1,0]];[prms.[2,0]]]
        K*Rt
    
    let hnorm(A:double Matrix) = 
        let partialResult = A |> Matrix.mapCols (fun _ col -> col/col.[col.Count-1])
        let res = partialResult.[0..partialResult.RowCount-2, 0..partialResult.ColumnCount-1]
        res

    let get_e P x3D x_measured = 
        hnorm (P*x3D) - x_measured

    let reshape (mIn:double Matrix) = 
        CreateVector.DenseOfArray(mIn.ToColumnMajorArray())
      
    let repmat_column n (vIn: double Vector) = 
        let mOut = DenseMatrix.zero<double> vIn.Count n 
        for i in 0..n-1 do
            mOut.SetColumn(i, vIn)
        mOut

    let repmat_row n (vIn: double Vector) = 
        let mOut = DenseMatrix.zero<double> n vIn.Count 
        for i in 0..n-1 do
            mOut.SetRow(i, vIn)
        mOut

    let to_mat (mam:double[,]) = 
        Matrix<double>.Build.DenseOfArray(mam)
   
    let get_J(prms:double Matrix, K:double Matrix, x3D: double Matrix) = 
        let a = prms.[3,0]
        let b = prms.[4,0]
        let c = prms.[5,0]
        let X = hnorm x3D

        let Rx = get_Rx a
        let Ry = get_Ry b
        let Rz = get_Rz c
        let P = get_P K prms

        let dRx = get_dRx a
        let dRy = get_dRy b
        let dRz = get_dRz c
        
        let xys = P*x3D
        let x = xys.Row(0)
        let y = xys.Row(1)
        let s = xys.Row(2)

        let df_dtx = (K * matrix[[1.0];[0.0];[0.0]]).ToRowMajorArray()
        let df_dty = (K * matrix[[0.0];[1.0];[0.0]]).ToRowMajorArray()
        let df_dtz = (K * matrix[[0.0];[0.0];[1.0]]).ToRowMajorArray()

        let df_da = K*Rz*Ry*dRx*X
        let df_db = K*Rz*dRy*Rx*X
        let df_dc = K*dRz*Ry*Rx*X

        //d(f/s)/d_anything
        let dfs_dtx = matrix [ df_dtx.[0]*s - x*df_dtx.[2];
                               df_dtx.[1]*s - y*df_dtx.[2] ]

        let dfs_dty = matrix [ df_dty.[0]*s - x*df_dty.[2];
                               df_dty.[1]*s - y*df_dty.[2] ]

        let dfs_dtz = matrix [ df_dtz.[0]*s - x*df_dtz.[2];
                               df_dtz.[1]*s - y*df_dtz.[2] ]

        let dfs_da = matrix [ df_da.Row(0).*s - x.*df_da.Row(2);
                              df_da.Row(1).*s - y.*df_da.Row(2) ]

        let dfs_db = matrix [ df_db.Row(0).*s - x.*df_db.Row(2);
                              df_db.Row(1).*s - y.*df_db.Row(2) ]

        let dfs_dc = matrix [ df_dc.Row(0).*s - x.*df_dc.Row(2);
                              df_dc.Row(1).*s - y.*df_dc.Row(2) ]
        
        //s^2
        let s_2 = s.*s |> repmat_row 2
                       |> reshape 
                       |> repmat_column 6
        
        (matrix [ reshape dfs_dtx;
                  reshape dfs_dty;
                  reshape dfs_dtz;
                  reshape dfs_da;
                  reshape dfs_db;
                  reshape dfs_dc ] |> Matrix.transpose)./s_2
                 
    let one_iter(params_in:double[], k_in:double[,], x3D_in:double[,], x_measured_in:double[,], cond:bool) = 
        let prms = Matrix<double>.Build.DenseOfColumnArrays params_in
        let x3D = to_mat x3D_in
        let x_measured = to_mat x_measured_in
        let K = to_mat k_in
        let P = get_P K prms
        let e = get_e P x3D x_measured |> reshape 
        let J = get_J(prms, K, x3D)
        let delta = -((Matrix.inverse ((Matrix.transpose J) * J)) * (Matrix.transpose J) * e.ToColumnMatrix())
        let prms_i1 = (prms + delta)
        if cond then
            System.Console.Write(e.L2Norm().ToString() + " ")
            //System.Console.WriteLine(" - L2 distance of detected and reprojected points")
            //System.Console.WriteLine(prms_i1)
        prms_i1.ToColumnMajorArray()

   
module matrixTransform = 
    let deg2rad a = a*(System.Math.PI / 180.0)

    let deg2rad_array (a:double[]) = [| for i in 0 .. 2 -> a.[i];
                                        for i in 3 .. 5 -> deg2rad a.[i] |]

    let get_T(prms_in:double[]) = 
        let prms = (*deg2rad_array*) prms_in
        let R = Matrix<double>.Build.DenseOfColumnArrays prms.[3..5] |> GaussNewton.get_R
        let t = Matrix<double>.Build.DenseOfColumnArrays prms.[0..2]
        let T = DenseMatrix.zero<double> 4 4
        T.[0..2,0..2] <- R
        T.[0..2,3..3] <- t
        T.[3..3,0..3] <- matrix[ [0.0; 0.0; 0.0; 1.0] ]
        T

    let transform_rbt_to_printer(params_printer2cam:double[], params_robot2cam:double[]) =
        let T_p_in_c = get_T params_printer2cam
        let T_r_in_c = get_T params_robot2cam
        let T_inv = Matrix.inverse T_r_in_c
        T_p_in_c*(Matrix.inverse T_r_in_c)

    let backprojection(parameters:double[], k:double[,], x3D:double[,]) = 
        let prms = Matrix<double>.Build.DenseOfColumnArrays parameters
        let K = GaussNewton.to_mat k
        let P = GaussNewton.get_P K prms
        let X = GaussNewton.to_mat x3D
        GaussNewton.hnorm(P*X).ToArray()

    let backproject_error(parameters:double[], k:double[,], x3D:double[,], x2D:double[,]) =
        let P = GaussNewton.get_P (GaussNewton.to_mat k) (Matrix<double>.Build.DenseOfColumnArrays parameters)
        let error = GaussNewton.get_e P (GaussNewton.to_mat x3D) (GaussNewton.to_mat x2D)
        error
        
    let decompose_transform_matrix(R:double Matrix) =
        let prms = Array.zeroCreate 6
        prms.[0..2] <- R.[0..2,3..3].ToColumnMajorArray()
        let sy = sqrt (R.[0,0]**2.0 + R.[1,0]**2.0)
        let a = atan2 R.[2,1] R.[2,2]
        let b = atan2 -R.[2,0] sy
        let c = atan2 R.[1,0] R.[0,0]
        prms.[3..5] <- [|a; b; c|]
        prms


      
   

