﻿namespace FSLibrary
open MathNet.Numerics.LinearAlgebra

module LevenbergMarquardt =
    open MathNet.Numerics.Random

    let get_exp(p:double[]) (x:double Matrix) = 
        (-p.[1]*(x-p.[2])).PointwiseExp()

    let sigmoid(p:double[]) (x:double Matrix) = 
        //y = L/(1+exp(-k*(x-x0))) + d
        let citatel = 1.0 + get_exp p x
        p.[0]/(citatel) + p.[3]

    let sigmoidCs(p:double[], x_in:double[]) =         
        let x = Matrix<double>.Build.DenseOfColumnArrays x_in
        let res = sigmoid p x
        res.ToColumnMajorArray()

    let get_e p (x:double Matrix) (y:double Matrix) =
        (sigmoid p x) - y

    let get_J(prms:double[]) (x:double Matrix) =
        let L = prms.[0]
        let k = prms.[1]
        let x0 = prms.[2]
        let d = prms.[3]
        let exp = get_exp prms x

        let df_dL = 1./(1.0+exp)
        let df_dk = L*(x-x0).*exp./(1.0+exp).PointwisePower(2.0)
        let df_dx0 = -L*k*exp./(1.0+exp).PointwisePower(2.0)
        let df_dd = DenseMatrix.init x.RowCount x.ColumnCount (fun _ _ -> double 1.0)
        matrix [GaussNewton.reshape df_dL;
                GaussNewton.reshape df_dk;
                GaussNewton.reshape df_dx0;
                GaussNewton.reshape df_dd] |> Matrix.transpose


    let one_iter(params_in:double[], x_in:double[], y:double[], lambda:double) = 
        let prms = Matrix<double>.Build.DenseOfColumnArrays params_in //L,k,x0,d
        let L = params_in.[0]
        let k = params_in.[1]
        let x0 = params_in.[2]
        let d = params_in.[3]
        let x = Matrix<double>.Build.DenseOfColumnArrays x_in
        let y = Matrix<double>.Build.DenseOfColumnArrays y
        let e = get_e params_in x y 
        let J = get_J params_in x
        let D = DiagonalMatrix.identity<double> 4
        let delta = -((Matrix.inverse ((Matrix.transpose J) * J + lambda * D)) * (Matrix.transpose J) * e)
        let prms_i1 = (prms + delta)
        prms_i1.ToColumnMajorArray()

    let get_initial_lambda(prms:double[], x:double[]) =
        let J = get_J prms (Matrix<double>.Build.DenseOfColumnArrays x)
        let H = ((Matrix.transpose J )* J).Diagonal()
        1e-3 * H.Sum() / double H.Count

