﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;
using ImageProcessing.DistributedAlgorithms.MarkerGenerating;
using MathNet.Numerics.Statistics;
using OpenCvSharp;
using YSoft.Rqa.VisualProwess.AutoCalibration.Db;
using YSoft.Rqa.VisualProwess.AutoCalibration.Db.Entities;
using YSoft.Rqa.VisualProwess.AutoCalibration.Exceptions;
using YSoft.Rqa.VisualProwess.AutoCalibration.MarkerDetecting;
using Point = OpenCvSharp.Point;

namespace YSoft.Rqa.VisualProwess.AutoCalibration.CameraCalibration
{
    public class CameraCalibrator: IDisposable
    {
        private readonly CalibrationContext _calibrationContext;

        public CameraCalibrator(CalibrationContext calibrationContext)
        {
            _calibrationContext = calibrationContext;
        }
        ///Method Calibrate() performs Camera calibration by detecting
        ///marker on the Device screen using set of rules written directly for one type of marker.
        /// <returns>Method returns 4 Point2f which represent 4 corners of screen</returns>
        ///
        ///Arguments:
        ///  <param name="deviceModelId">ID of used Device</param>
        ///  <param name="screenSize">Size of the Device screen</param>,
        ///  <param name="img">Image of class Mat of MatType.CV_8UC3 preferably, but should manage all C3 MatTypes</param>.
        ///
        ///Exceptions:
        ///  <exception cref="ArgumentNullException">thrown if Mat image is Null</exception>
        ///  <exception cref="MarkerCannotBeDetectedException"></exception>

        public async Task<List<Point2f>> Calibrate(int deviceModelId, Size screenSize, Mat img, List<Point2f> currentCalibration = null)
        {
            var device = await Device.GetById(deviceModelId, _calibrationContext);
            var marker = new MarkerDetector(device.GetMarkerTemplate, img).Detect();
            //var img2 = img.Clone();
            //for (int i = 0; i < 4; i++)
            //{
            //    Cv2.DrawContours(img2, new List<List<Point>> { marker.CornerShapes[i].Contour.AllPoints }, 0,
            //        new Scalar(0, 255, 0), 1);
            //    for (int j = 0; j < 5; j++)
            //    {
            //        Cv2.Circle(img2, (Point)marker.CornerShapes[i].AllPoints[j], 5, Scalar.Red);
            //    }
            //}
            //Cv2.ImShow("marker", img2);
            //Cv2.WaitKey(0);
            //Cv2.ImWrite("camCalibMarker.png", img2);
            
            var approximatedLines = new List<Line2D>();

            foreach (ScreenSide side in Enum.GetValues(typeof(ScreenSide)))
            {
                var sidePoints = GetSidePoints(side, marker);

                approximatedLines.Add(Cv2.FitLine(sidePoints.Select(x=> new Point(x.X, x.Y)), DistanceTypes.L2, 0, .01, .01));
                //DrawLineAndPoints(approximatedLines.Last(), sidePoints, side, img);
                //Cv2.WaitKey();
            }
            DrawLines(approximatedLines, img);
            Cv2.WaitKey(0);
            var intersections = CalculateIntersectionsOfLines(approximatedLines);
            var points = AdjustIntersectionsByMarkerOffset(intersections, screenSize, marker.MarkerTemplate.Offset);
            var rotatedPoints = points.ConvertToPoint2f().ToList().RotateList(1);
            ValidateCalibration(rotatedPoints, currentCalibration);
            DrawPoints(points, img);
            Cv2.WaitKey(0);

            return rotatedPoints;
        }

        public static void ValidateCalibration(List<Point2f> points, List<Point2f> currentCalibration) {
            // if bottom points are lower in Y axis than upper points
            if (points[0].Y < points[1].Y || points[3].Y < points[2].Y) {
                throw new ValidationCalibrationException($"Y Axis does not seem valid. {string.Join(", ", points.Select(x => x.Y))}");
            }
            // if right points are not in right of left points
            if (points[2].X < points[1].X || points[2].X < points[0].X || points[3].X < points[1].X || points[3].X < points[0].X) {
                throw new ValidationCalibrationException($"X Axis does not seem valid. {string.Join(", ", points.Select(x => x.X))}");
            }

            // if LD point is closer to RU (diagonal) than LU or RD
            if (points[0].DistanceTo(points[1])>points[0].DistanceTo(points[2]) || points[0].DistanceTo(points[3]) > points[0].DistanceTo(points[2])) {
                throw new ValidationCalibrationException($"Left lower point does not seem valid.");
            }
            // if LU point is closer to RD (diagonal) than LD or RU
            if (points[1].DistanceTo(points[0]) > points[1].DistanceTo(points[3]) || points[1].DistanceTo(points[2]) > points[1].DistanceTo(points[3])) {
                throw new ValidationCalibrationException($"Left upper point does not seem valid.");
            }
            // if RU point is closer to LD (diagonal) than LU or RD
            if (points[2].DistanceTo(points[1]) > points[2].DistanceTo(points[0]) || points[2].DistanceTo(points[3]) > points[2].DistanceTo(points[0])) {
                throw new ValidationCalibrationException($"Right upper point does not seem valid.");
            }
            // if RD point is closer to LU (diagonal) than LD or RU
            if (points[3].DistanceTo(points[0]) > points[3].DistanceTo(points[1]) || points[3].DistanceTo(points[2]) > points[3].DistanceTo(points[1])) {
                throw new ValidationCalibrationException($"Right lower point does not seem valid.");
            }

            if (currentCalibration != null)
            {
                var dist = new List<double>() {
                    points[0].DistanceTo(currentCalibration[0]), points[1].DistanceTo(currentCalibration[1]),
                    points[2].DistanceTo(currentCalibration[2]), points[3].DistanceTo(currentCalibration[3])
                };

                for (int i = 0; i < 4; i++) {
                    if (dist[i] > 3 * dist.Where((v, j) => j != i).ToList().Mean()) {
                        throw new ValidationCalibrationException($"Point {i} is significantly deviated from its previous calibration position compared to other points positions.");
                    }
                }
            }
        }

        private List<Point2d> GetSidePoints(ScreenSide side, Marker marker)
        {
            var perpendicularSides = Helpers.GetPerpendicularSides(side);
            var res = marker.CornerShapes
                    .Where(c => c.Localization.ToString().Contains(side.ToString()))
                    .SelectMany(c => new List<Point2d>
                        {
                            c.TopPoints[(perpendicularSides[0]+1)%4],
                            c.TopPoints[(perpendicularSides[1]+1)%4],
                            c.CenterPoint
                        }
                       )
                    .ToList();
            return res;
        }
        
        private Point2d[] CalculateIntersectionsOfLines(List<Line2D> approximatedLines)
        {
            // sides are in order: R D L U
            // corners are in order: RD, LD, LU, RU
            int n = approximatedLines.Count < 4 ? approximatedLines.Count-1 : 4;
            var points = new Point2d[n];
            for (int i = 0; i < n; i++)
            {
                points[i] = Helpers.CalculateIntersection(approximatedLines[i], approximatedLines[(i + 1) % 4]);
            }
            return points;
        }

        private Point2d[] AdjustIntersectionsByMarkerOffset(Point2d[] intersections, Size screenSize, int offset)
        {
            var screenModel = new Point2d[4]
            {
                new Point2d(screenSize.Width, screenSize.Height),
                new Point2d(0, screenSize.Height),
                new Point2d(0, 0),
                new Point2d(screenSize.Width, 0)
            };
            var intersectionsModel = new Point2d[4]
            {
                new Point2d(screenSize.Width + offset, screenSize.Height + offset),
                new Point2d(-offset, screenSize.Height + offset),
                new Point2d(-offset,-offset),
                new Point2d(screenSize.Width + offset, -offset)
            };
            var transformMatrixModelToReal = Cv2.GetPerspectiveTransform(
                intersectionsModel.ConvertToPoint2f(), intersections.ConvertToPoint2f());
            var transformedPoints =  Cv2.PerspectiveTransform(screenModel, transformMatrixModelToReal);
            return transformedPoints;
        }
        
        private void DrawLine(Line2D line, ScreenSide side, Mat image)
        {
            var img = image.Clone();
            int coef = 1000;
            Cv2.Line(img, new Point(line.X1+coef*line.Vx , line.Y1+coef*line.Vy),
                new Point(line.X1 - coef * line.Vx, line.Y1 -coef * line.Vy),new Scalar(255,100,10));
            Cv2.ImShow("line" + side, img);
        }

        private void DrawLineAndPoints(Line2D line, List<Point2d> points, ScreenSide side, Mat image)
        {
            var img = image.Clone();
            int coef = 1000;
            Cv2.Line(img, new Point(line.X1 + coef * line.Vx, line.Y1 + coef * line.Vy),
                new Point(line.X1 - coef * line.Vx, line.Y1 - coef * line.Vy), new Scalar(255, 100, 10));
            foreach (var point in points)
            {
                Cv2.Circle(img, new Point(point.X, point.Y), 3, new Scalar(10, 200, 50));
            }
            Cv2.ImShow("SidePoints" + side.ToString(), img);
            //Cv2.ImWrite("camCalibSidePoints" + side.ToString() + ".png", img);
        }

        private void DrawLines(List<Line2D> lines, Mat image)
        {
            var img = image.Clone();
            int coef = 1000;
            foreach (var line in lines)
            {
                Cv2.Line(img, new Point(line.X1 + coef * line.Vx, line.Y1 + coef * line.Vy),
                    new Point(line.X1 - coef * line.Vx, line.Y1 - coef * line.Vy), new Scalar(255, 100, 10), 2);
            }
            Cv2.ImShow("line", img);
            //Cv2.ImWrite("camCalibLines.png", img);
        }

        private void DrawSidePoints(List<Point2d> points, ScreenSide side, Mat image)
        {
            var img = image.Clone();
            foreach (var point in points)
            {
                Cv2.Circle(img, new Point(point.X, point.Y), 3, new Scalar(10, 200, 50));
            }
            Cv2.ImShow("SidePoints" + side.ToString(), img);
            //Cv2.ImWrite("camCalibSidePoints" + side.ToString() + ".png", img);
        }

        private void DrawPoints(List<Point> points, ScreenCorner corner, Mat image)
        {
            var img = image.Clone();
            foreach (var point in points)
            {
                Cv2.Circle(img, point, 3, new Scalar(10, 200, 50));
            }
            Cv2.ImShow("Points "+corner, img);
        }

        private void DrawPoints(List<Point2d> points, Mat image)
        {
            var img = image.Clone();
            foreach (var point in points)
            {
                Cv2.Circle(img, new Point(point.X, point.Y), 3, new Scalar(10, 200, 50));
            }
            Cv2.ImShow("Points", img);
        }

        private void DrawPoints(Point2d[] points, Mat image)
        {
            var img = image.Clone();
            foreach (var point in points)
            {
                Cv2.Circle(img, new Point(point.X, point.Y), 3, new Scalar(10, 200, 50), 3);
            }
            Cv2.ImShow("Points", img);
            //Cv2.ImWrite("camCalibPoints.png", img);

        }

        public void Dispose()
        {
            _calibrationContext.Dispose();
        }
    }
}