﻿using System;
using System.Collections.Generic;
using System.Linq;
using FSLibrary;
using ImageProcessing.DistributedAlgorithms.MarkerGenerating;
using OpenCvSharp;
using YSoft.Rqa.VisualProwess.AutoCalibration.MarkerDetecting;

namespace YSoft.Rqa.VisualProwess.AutoCalibration.DetectionRefine
{
    internal class DetectionRefiner
    {
        private CornerShape Shape { get; set; }
        private Mat Image { get; set; }

        private const int Offset = 7;
        private readonly double[] _initialEstimation = { 0.54, 0.14, 0.80, 0.89 }; //L,k,x0,d
        private const int MaxIter = 400;
        private const int Factor = 2;
        private const double MinChange = 1e-7;
        private const double TakeFirstSigmoidsInPercent = 80;

        internal DetectionRefiner(CornerShape shape, Mat image)
        {
            Shape = shape;
            Image = image.CvtColor(ColorConversionCodes.RGB2GRAY, 1);
            Image.ConvertTo(Image, MatType.CV_64FC1, 1.0/255.0, 0);
        }

        internal void RefineShape(bool manualControl = false)
        {
            var lines = new List<Line2D>();
            
            for (int i = 0; i < Shape.AllPoints.Count; i++)
            {
                var start = Shape.AllPoints[i];
                var stop = Shape.AllPoints[(i + 1) % Shape.AllPoints.Count];
                var img = ConvertImg();
                Cv2.Line(img, (int)start.X, (int)start.Y, (int)stop.X, (int)stop.Y, Scalar.Black);

                var k = (start.Y - stop.Y)/(start.X - stop.X);
                var q = start.Y - k * start.X;

                var inflexPoints = new List<Point2f>();
                var sigmoids = new List<Sigmoid>();

                //horizontal --
                if (Math.Abs(k) < 0.3) 
                {
                    var yStart = (int)Math.Min(start.Y, stop.Y) - Offset;
                    var yStop = (int)Math.Max(start.Y, stop.Y) + Offset;
                    var xStart = (int)Math.Min(start.X, stop.X);
                    var xStop = (int)Math.Max(start.X, stop.X);

                    for (int x = xStart; x < xStop; x++)
                    {
                        var mask = Image.SubMat(yStart, yStop+1, x, x + 1);
                        double[] yArray = new double[mask.Rows * mask.Cols];
                        mask.GetArray(0, 0, yArray);
                        var xArray = Enumerable.Range(0, yArray.Length).ToArray();
                        var prms = LevenbergMarquardtFs(xArray, yArray, _initialEstimation); //L,k,x0,d
                        var x0 = prms[2];
                        var inflexPoint = new Point2f(x,yStart + (float)x0);
                        sigmoids.Add(
                            new Sigmoid(xArray, yArray,prms,inflexPoint)
                            );

                        if (manualControl)
                        {
                            //var img = ConvertImg();
                            Cv2.Line(img, (int) start.X, (int) start.Y, (int) stop.X, (int) stop.Y, Scalar.Black);
                            Cv2.Line(img, x, yStart, x, yStop, Scalar.White);
                            Cv2.ImShow("control", img);
                            Cv2.WaitKey();
                        }
                    }
                }
                //vertical |
                else if (k < -3 || k > 3)
                {
                    var yStart = (int)Math.Min(start.Y, stop.Y);
                    var yStop = (int)Math.Max(start.Y, stop.Y);
                    var xStart = (int)Math.Min(start.X, stop.X) - Offset;
                    var xStop = (int)Math.Max(start.X, stop.X) + Offset;

                    for (int y = yStart; y < yStop; y++)
                    {
                        var mask = Image.SubMat(y, y + 1, xStart, xStop+1);
                        double[] yArray = new double[mask.Rows * mask.Cols];
                        mask.GetArray(0, 0, yArray);
                        var xArray = Enumerable.Range(0, yArray.Length).ToArray();
                        var prms = LevenbergMarquardtFs(xArray, yArray, _initialEstimation); //L,k,x0,d

                        var x0 = prms[2];
                        var inflexPoint = new Point2f(xStart+(float)x0, y);
                        sigmoids.Add(
                            new Sigmoid(xArray, yArray, prms, inflexPoint)
                        );

                        if (manualControl)
                        {
                            //var img = ConvertImg();
                            Cv2.Line(img, (int) start.X, (int) start.Y, (int) stop.X, (int) stop.Y, Scalar.Black);
                            Cv2.Line(img, xStart, y, xStop, y, Scalar.White);
                            Cv2.ImShow("control", img);
                            Cv2.WaitKey();
                        }
                    }
                }
                //major diagonal \
                else if (k >= 0.3 && k <= 3) 
                {
                    var xStart = (int)Math.Round(Math.Min(start.X, stop.X));
                    var xStop = (int)Math.Round(Math.Max(start.X, stop.X));

                    for (int x = xStart; x < xStop; x++)
                    {
                        int y = (int)(k * x + q);
                        var mask = Image.SubMat(y-Offset, y+Offset+1, x-Offset, x+Offset+1);
                        var diagonal = mask.Transpose().Flip(FlipMode.Y).Diag(MatDiagType.Main);
                        double[] yArray = new double[diagonal.Rows * diagonal.Cols];
                        diagonal.GetArray(0, 0, yArray);
                        var xArray = Enumerable.Range(0, yArray.Length).ToArray();
                        var prms = LevenbergMarquardtFs(xArray, yArray, _initialEstimation); //L,k,x0,d
                         var x0 = prms[2];
                        var inflexPoint = new Point2f((float)(x-Offset+x0), (float)(y+Offset-x0));
                        sigmoids.Add(
                            new Sigmoid(xArray, yArray, prms, inflexPoint)
                        );

                        if (manualControl)
                        {
                            //var img = ConvertImg();
                            Cv2.Line(img, (int) start.X, (int) start.Y, (int) stop.X, (int) stop.Y, Scalar.Black);
                            Cv2.Line(img, x - Offset, y + Offset, x + Offset, y - Offset, Scalar.White);
                            Cv2.ImShow("control", img);
                            Cv2.WaitKey();
                        }
                    }
                }
                //minor diagonal /
                else if (k <= -0.3 && k >= -3) 
                {
                    var xStart = (int)Math.Round(Math.Min(start.X, stop.X));
                    var xStop = (int)Math.Round(Math.Max(start.X, stop.X));

                    for (int x = xStart; x < xStop; x++)
                    {
                        int y = (int)(k * x + q);
                        var mask = Image.SubMat(y - Offset, y + Offset + 1, x - Offset, x + Offset + 1);
                        var diagonal = mask.Diag(MatDiagType.Main);
                        double[] yArray = new double[diagonal.Rows * diagonal.Cols];
                        diagonal.GetArray(0, 0, yArray);
                        var xArray = Enumerable.Range(0, yArray.Length).ToArray();
                        var prms = LevenbergMarquardtFs(xArray, yArray, _initialEstimation); //L,k,x0,d
                        var x0 = prms[2];
                        var inflexPoint = new Point2f((float)(x - Offset + x0), (float)(y - Offset + x0));
                        sigmoids.Add(
                            new Sigmoid(xArray, yArray, prms, inflexPoint)
                        );

                        if (manualControl)
                        {
                            //var img = ConvertImg();
                            Cv2.Line(img, (int) start.X, (int) start.Y, (int) stop.X, (int) stop.Y, Scalar.Black);
                            Cv2.Line(img, x - Offset, y - Offset, x + Offset, y + Offset, Scalar.White);
                            Cv2.ImShow("control", img);
                            Cv2.WaitKey();
                        }
                    }
                }

                var rmse = sigmoids.Select(x => x.GetRmse()).ToList();
                rmse.Sort();

                  lines.Add(
                    RefinedEdge.Interpolate(sigmoids, TakeFirstSigmoidsInPercent)
                );
            }

            lines = lines.RotateList(lines.Count - 1);
            //DrawLines(lines);

            var points = new List<Point2d>();
            for (int i = 0; i < lines.Count; i++)
            {
                points.Add(Helpers.CalculateIntersection(lines[i], lines[(i+1) % lines.Count]));
            }
            Shape.AllPoints = new List<Point2d>(points);
            var centerPoint = CornerShape.ReorganizePoints(points);
            Shape.CenterPoint = centerPoint;
            Shape.TopPoints = points;
        }

        internal static double[] LevenbergMarquardtFs(int[] xArray, double[] yArray, double[] initEstim)
        {
            var x = xArray.Select(i => (double) i).ToArray();
            var y = yArray;
            var params_i = new double[4];
            initEstim.CopyTo(params_i,0);
            params_i[2] = (x[0] + x[x.Length - 1])/2;
            var params_i1 = new double[]{1000,1000,1000,1000};
            var iter = 0;
            var lambda = LevenbergMarquardt.get_initial_lambda(params_i, x);
            var error_i = LevenbergMarquardt.sigmoidCs(params_i, x)
                .Zip(y, (a,b) => a-b)
                .ToArray();
            var norm = 100d;

            while (iter < MaxIter && norm > MinChange)
            {
                iter++;
                params_i1 = LevenbergMarquardt.one_iter(params_i, x, y, lambda);
                var error_i1 = LevenbergMarquardt.sigmoidCs(params_i1, x)
                    .Zip(y, (a,b) => a-b)
                    .ToArray();
                norm = Helpers.NormL2(params_i, params_i1);

                if (error_i1.Select(j => j * j).Sum() < error_i.Select(j => j * j).Sum())
                {
                    lambda /= Factor;
                    params_i1.CopyTo(params_i, 0);
                    error_i1.CopyTo(error_i, 0);
                }
                else lambda *= Factor;
            }
            return params_i1;
        }
        
        private void DrawLines(List<Line2D> lines)
        {
            var img = ConvertImg();
            int lineKoef = 1000;
            foreach (var line in lines)
            {
                Cv2.Line(img, new Point(line.X1 + lineKoef * line.Vx, line.Y1 + lineKoef * line.Vy),
                    new Point(line.X1 - lineKoef * line.Vx, line.Y1 - lineKoef * line.Vy), 255);
            }
            Cv2.ImShow("lines",img);
            Cv2.WaitKey();
        }

        private Mat ConvertImg()
        {
            var img = new Mat();
            Image.Clone().ConvertTo(img, MatType.CV_8UC1, 255);
            return img;
        }
    }
}