﻿using System;
using System.Collections.Generic;
using System.Data;
using System.IO;
using System.Linq;
using System.Threading.Tasks;
using FSLibrary;
using MathNet.Numerics.LinearAlgebra;
using OpenCvSharp;
using YSoft.Rqa.VisualProwess.AutoCalibration.Db;
using YSoft.Rqa.VisualProwess.AutoCalibration.Db.Entities;
using YSoft.Rqa.VisualProwess.AutoCalibration.MarkerDetecting;
using YSoft.Rqa.VisualProwess.AutoCalibration.RobotCalibration;

namespace YSoft.Rqa.VisualProwess.AutoCalibration.DiplomkaTests
{
    public static class MarkerChessTester
    {
        public static async Task<Matrix<double>> Test(int devId, int camId, int rbtId, Size screenSize, Mat img,
            CalibrationContext context, bool rotateChessPoints = false)
        {
            //with distortion
            var rbtCalibrator = new RobotCalibrator(context);
            await rbtCalibrator.Init(devId, screenSize, rbtId, camId, img);
            var paramsChess = rbtCalibrator.RobotToCamera(rotateChessPoints);
            var paramsMarker = rbtCalibrator.ScreenToCamera();

            var markerTemplate = rbtCalibrator.Device.GetMarkerTemplate;
            var marker = new MarkerDetector(markerTemplate, img).Detect();
            //var res = await rbtCalib.Calibrate(3, new Size(198, 111), 1, 1, im);
            var error = matrixTransform.backproject_error(paramsMarker, rbtCalibrator.Camera.GetIntrinsicMatrix(),
                rbtCalibrator.Device.GetMarkerTemplate.GetMarkerModel(screenSize),
                rbtCalibrator.ExtractAllPlanePoints(marker));

            var mat = matrixTransform.transform_rbt_to_printer(paramsMarker, paramsChess);
            Console.WriteLine(mat);
            return mat;
        }

        public static async Task<Matrix<double>> UndistortTest(int devId, int camId, int rbtId, Size screenSize, Mat img,
            CalibrationContext context, bool rotateChessPoints = false)
        {
            //without distortion
            var cam = await Camera.GetById(camId, context);
            var imageUndist = cam.UndistortImage(img);

            var rbtCalibrator = new RobotCalibrator(context);
            await rbtCalibrator.Init(devId, screenSize, rbtId, camId, imageUndist);
            var paramsUndist = rbtCalibrator.RobotToCamera(rotateChessPoints);
            var paramsMarker = rbtCalibrator.ScreenToCamera();

            var markerTemplate = rbtCalibrator.Device.GetMarkerTemplate;
            var marker = new MarkerDetector(markerTemplate, img).Detect();
            var error = matrixTransform.backproject_error(paramsMarker, rbtCalibrator.Camera.GetIntrinsicMatrix(),
                rbtCalibrator.Device.GetMarkerTemplate.GetMarkerModel(screenSize),
                rbtCalibrator.ExtractAllPlanePoints(marker));

            var matUndist = matrixTransform.transform_rbt_to_printer(paramsMarker, paramsUndist);
            return matUndist;
        }

        public static async ValueTask<Matrix<double>> PnPTest(int devId, int camId, int rbtId, Size screenSize, Mat img,
            CalibrationContext context, bool rotateChessPoints = false)
        {
            //PnP
            var cam = await Camera.GetById(camId, context);
            var imageUndist = cam.UndistortImage(img);

            var rbtCalibrator = new RobotCalibrator(context);
            await rbtCalibrator.Init(devId, screenSize, rbtId, camId, imageUndist);
            var paramsChess = rbtCalibrator.SolveChessByPnP(rotateChessPoints);
            var paramsMarker = new double[6];

            var matPnP = matrixTransform.transform_rbt_to_printer(paramsMarker, paramsChess);
            return matPnP;
        }

        public static void GenerateExcelAllTest(double[] origPrms, List<double[]> estimatePrms)
        {
            var list = new List<double[]>(estimatePrms);
            list.Insert(0,origPrms);
            var table = GetDataTableAllTest(list);
            var lines = new List<string>();

            string[] columnNames = table.Columns.Cast<DataColumn>().
                Select(column => column.ColumnName).
                ToArray();

            var header = string.Join(",", columnNames);
            lines.Add(header);

            var valueLines = table.AsEnumerable()
                .Select(row => string.Join(",", row.ItemArray));
            lines.AddRange(valueLines);

            File.WriteAllLines(@"..\..\..\" + Path.GetRandomFileName() + ".csv", lines);
        }

        public static System.Data.DataTable GetDataTableAllTest(List<double[]> prms)
        {
            System.Data.DataTable table = new System.Data.DataTable();
            table.Columns.Add("test", typeof(int));
            table.Columns.Add("$t_x$ [mm]", typeof(double));
            table.Columns.Add("$t_y$ [mm]", typeof(double));
            table.Columns.Add("$t_z$ [mm]", typeof(double));
            table.Columns.Add(@"$\alpha [rad]$", typeof(double));
            table.Columns.Add(@"$\beta [rad]$", typeof(double));
            table.Columns.Add(@"$\gamma [rad]$", typeof(double));
            
            for (int i = 0; i < prms.Count; i++)
            {
                var row = prms[i].ToList();
                row.Insert(0, i);
                var dataRow = table.NewRow();
                for (int j = 0; j < table.Columns.Count; j++)
                {
                    dataRow[j] = row[j];
                }
                table.Rows.Add(dataRow);
            }
            return table;
        }

        public static async Task<double[]> ChangeKTest(int devId, int camId, int rbtId, Size screenSize, Mat img,
            CalibrationContext context, double[] changeK, bool rotateChessPoints = false)
        {
            var cam = await Camera.GetById(camId, context);
            var imageUndist = cam.UndistortImage(img);

            var rbtCalibrator = new RobotCalibrator(context);
            await rbtCalibrator.Init(devId, screenSize, rbtId, camId, imageUndist);
            rbtCalibrator.Camera.ChangeK(changeK[0], changeK[1], changeK[2], changeK[3]);
            var paramsUndist = rbtCalibrator.RobotToCamera(rotateChessPoints);
            var paramsMarker = rbtCalibrator.ScreenToCamera();

            var matUndist = matrixTransform.transform_rbt_to_printer(paramsMarker, paramsUndist);
            rbtCalibrator.Camera.ChangeK(-changeK[0], -changeK[1], -changeK[2], -changeK[3]);
            return matrixTransform.decompose_transform_matrix(matUndist);
        }

        public static void GenerateExcelCFshift(List<double[]> shift, List<double[]> estimatePrms)
        {
            var table = GetDataTableCFshift(shift, estimatePrms);
            var lines = new List<string>();

            string[] columnNames = table.Columns.Cast<DataColumn>().
                Select(column => column.ColumnName).
                ToArray();

            var header = string.Join(",", columnNames);
            lines.Add(header);

            var valueLines = table.AsEnumerable()
                .Select(row => string.Join(",", row.ItemArray));
            lines.AddRange(valueLines);

            File.WriteAllLines(@"..\..\..\" + Path.GetRandomFileName() + ".csv", lines);
        }

        public static System.Data.DataTable GetDataTableCFshift(List<double[]> shifts, List<double[]> prms)
        {
            System.Data.DataTable table = new System.Data.DataTable();
            table.Columns.Add("shift", typeof(string));
            table.Columns.Add("$t_x$ [mm]", typeof(double));
            table.Columns.Add("$t_y$ [mm]", typeof(double));
            table.Columns.Add("$t_z$ [mm]", typeof(double));
            table.Columns.Add(@"$\alpha [rad]$", typeof(double));
            table.Columns.Add(@"$\beta [rad]$", typeof(double));
            table.Columns.Add(@"$\gamma [rad]$", typeof(double));

            for (int i = 0; i < prms.Count; i++)
            {
                var dataRow = table.NewRow();
                var shift = shifts[i];
                dataRow[0] = $"[{shift[0]};{shift[1]};{shift[2]};{shift[3]}]";
                for (int j = 1; j < table.Columns.Count; j++)
                {
                    dataRow[j] = prms[i][j-1];
                }
                table.Rows.Add(dataRow);
            }
            return table;
        }

        public static void GenerateExcelDifferentIntrinsic(List<double[]> prms, int[] ids)
        {
            var table = GetDataTableDifferentIntrinsic(prms, ids);
            var lines = new List<string>();

            string[] columnNames = table.Columns.Cast<DataColumn>().
                Select(column => column.ColumnName).
                ToArray();

            var header = string.Join(",", columnNames);
            lines.Add(header);

            var valueLines = table.AsEnumerable()
                .Select(row => string.Join(",", row.ItemArray));
            lines.AddRange(valueLines);

            File.WriteAllLines(@"..\..\..\" + Path.GetRandomFileName() + ".csv", lines);
        }

        public static System.Data.DataTable GetDataTableDifferentIntrinsic(List<double[]> prms, int[] ids)
        {
            System.Data.DataTable table = new System.Data.DataTable();
            table.Columns.Add("camera ID", typeof(int));
            table.Columns.Add("$t_x$ [mm]", typeof(double));
            table.Columns.Add("$t_y$ [mm]", typeof(double));
            table.Columns.Add("$t_z$ [mm]", typeof(double));
            table.Columns.Add(@"$\alpha$ [rad]", typeof(double));
            table.Columns.Add(@"$\beta$ [rad]", typeof(double));
            table.Columns.Add(@"$\gamma$ [rad]", typeof(double));

            for (int i = 0; i < prms.Count; i++)
            {
                var dataRow = table.NewRow();
                dataRow[0] = ids[i];
                for (int j = 1; j < table.Columns.Count; j++)
                {
                    dataRow[j] = prms[i][j-1];
                }
                table.Rows.Add(dataRow);
            }
            return table;
        }
    }
}
