﻿using System;
using System.Collections.Generic;
using System.Linq;
using ImageProcessing.DistributedAlgorithms.MarkerGenerating;
using OpenCvSharp;

namespace YSoft.Rqa.VisualProwess.AutoCalibration.MarkerDetecting
{
    internal class ContourForEvaluation
    {
        internal List<Point> AllPoints;
        //in Right Down Left Up order with Center somewhere
        internal List<Point> TopPoints { get; set; }
        internal Point CenterPoint {get; set; }
        internal List<int> TopPointsIndexes;
        internal int CenterPointIndex;
        //[RightDown, LeftDown,LeftUp,RightUp] with Center somewhere
        internal List<Line2D> Lines { get; set; }
        internal List<double> Rmse;

        internal ContourForEvaluation(List<Point> contour)
        {
            AllPoints = contour;
        }

        internal void Init()
        {
            var polygonPoints = Cv2.ApproxPolyDP(AllPoints, Helpers.ApproxPolyCoef, true).ToList();
            TopPoints = RotatePolygonPoints(polygonPoints);
            CenterPoint = GetCenterPoint();
            TopPointsIndexes = GetIndexes(out var centerIndex, out var rotatedContour);
            AllPoints = rotatedContour;
            CenterPointIndex = centerIndex;
            Lines = FitLines();
            Rmse = CalculateRmseOfContour();
        }

        private List<Point> RotatePolygonPoints(List<Point> polygonPoints)
        {
            //RDLU
            var offsetOfRightPoint = polygonPoints.IndexOf(polygonPoints.OrderByDescending(p => p.X).First());
            return polygonPoints.RotateList(offsetOfRightPoint);
        }

        private Point GetCenterPoint()
        {
            return TopPoints.Find(i =>
                i.X < TopPoints.Max(x => x.X)
                &&
                i.X > TopPoints.Min(x => x.X)
                &&
                i.Y < TopPoints.Max(x => x.Y)
                &&
                i.Y > TopPoints.Min(x => x.Y)
            );
        }

        private List<int> GetIndexes(out int centerIndex, out List<Point> rotatedContour)
        {
            var indexes = TopPoints.Select(p => AllPoints.IndexOf(p)).ToList();
            rotatedContour = indexes[0] == 0 ? AllPoints : AllPoints.RotateList(indexes[0]);
            centerIndex = rotatedContour.IndexOf(CenterPoint);
            return TopPoints.Select(p => AllPoints.RotateList(indexes[0]).IndexOf(p)).ToList();
        }

        private List<Line2D> FitLines()
        {
            var approximatedLines = new List<Line2D>();
            var fitLineIndexes = TopPointsIndexes.ToList();
            fitLineIndexes.Add(TopPointsIndexes[0]);

            for (int i = 0; i < TopPointsIndexes.Count; i++)
            {
                var index1 = fitLineIndexes[i];
                var index2 = fitLineIndexes[i + 1];
                if (index1 < index2)
                {
                    approximatedLines.Add(
                        Cv2.FitLine(
                            AllPoints.GetRange(index1, index2 - index1 + 1),
                            DistanceTypes.L2, 0, .01, .01)
                    );
                }
                else
                {
                    var p = AllPoints.Take(index2 + 1)
                        .Concat(AllPoints
                            .Skip(index1 - index2 - 1)
                            .Take(AllPoints.Count - index1)
                        ).ToList();
                    approximatedLines.Add(
                        Cv2.FitLine(p, DistanceTypes.L2, 0, .01, .01)
                    );
                }
            }
            return approximatedLines;
        }

        private List<double> CalculateRmseOfContour()
        {
            var rmse = new List<double>();
            var contour = AllPoints.ToList();
            contour.Add(contour.First());
            var indexes = TopPointsIndexes.ToList();
            indexes.Add(contour.Count - 1);
            for (int i = 0; i < TopPointsIndexes.Count; i++)
            {
                var contourSection = contour.Skip(indexes[i]).Take(indexes[i + 1] - indexes[i] + 1).ToList();
                rmse.Add(Math.Pow(
                    contourSection.Sum(point => Math.Pow(Lines[i].Distance(point), 2))
                    / contourSection.Count,
                    .5)
                );
            }
            return rmse;
        }

        internal ScreenCorner GetLocalizationOfCenterPoint()
        {
            switch(TopPoints.IndexOf(CenterPoint))
            {
                case 1:
                    return ScreenCorner.RightDown;
                case 2:
                    return ScreenCorner.LeftDown;
                case 3:
                    return ScreenCorner.LeftUp;
                case 4:
                    return ScreenCorner.RightUp;
                default:
                    throw new ArgumentOutOfRangeException("CenterPoint cannot be on another position in TopPoints list than [1,2,3 or 4]");
            }
        }
    }
}