﻿using System;
using System.Collections.Generic;
using System.Linq;
using ImageProcessing.DistributedAlgorithms.MarkerGenerating;
using OpenCvSharp;

namespace YSoft.Rqa.VisualProwess.AutoCalibration.MarkerDetecting
{
    internal class CornerShape : Shape
    {
        internal ScreenCorner Localization { get; set; }
        internal List<Point2d> TopPoints { get; set; }
        internal List<Point2d> AllPoints { get; set; }
        internal Point2d CenterPoint { get; set; }
        internal Mat Image { get; set; }

        internal CornerShape(ContourForEvaluation contour, ScreenCorner localization, Mat img) : base(contour)
        {
            Image = img.Clone();
            Localization = localization;
        }

        internal void Init()
        {
            TopPoints = GetTopPoints(base.Contour.AllPoints, out var centerPoint);
            CenterPoint = centerPoint;
        }

        private List<Point2d> GetTopPoints(List<Point> points, out Point2d centerPoint)
        {
            var polygonPoints = Cv2.ApproxPolyDP(points, Helpers.ApproxPolyCoef, true).ToList();
            var topContourPoints = RotatePolygonPoints(polygonPoints);
            var lines = FitLines(topContourPoints, points);
            var intersections = CalculateIntersections(lines);
            AllPoints = new List<Point2d>(intersections);
            centerPoint = ReorganizePoints(intersections);
            //Draw(lines, intersections, centerPoint, polygonPoints);
            return intersections;
        }

        internal static Point2d ReorganizePoints(List<Point2d> intersections)
        {
            var center = intersections.First(i =>
                i.X < intersections.Max(x => x.X)
                &&
                i.X > intersections.Min(x => x.X)
                &&
                i.Y < intersections.Max(x => x.Y)
                &&
                i.Y > intersections.Min(x => x.Y)
            );

            intersections.Remove(center);
            return center;
        }

        private Point[] RotatePolygonPoints(List<Point> polygonPoints)
        {
            //RDLU
            var offsetOfRightPoint = polygonPoints.IndexOf(polygonPoints.OrderByDescending(p => p.X).First());
            return polygonPoints.RotateList(offsetOfRightPoint).ToArray();
        }

        private List<Line2D> FitLines(Point[] topContourPoints, List<Point> points)
        {
            var approximatedLines = new List<Line2D>();
            var fitLinePoints = topContourPoints;
            Array.Resize(ref fitLinePoints, fitLinePoints.Length + 1);
            fitLinePoints[fitLinePoints.Length - 1] = fitLinePoints[0];

            for (int i = 0; i < topContourPoints.Length; i++)
            {
                var index1 = points.IndexOf(fitLinePoints[i]);
                var index2 = points.IndexOf(fitLinePoints[i + 1]);
                if (index1 < index2)
                {
                    approximatedLines.Add(
                        Cv2.FitLine(
                            points.GetRange(index1, index2 - index1 + 1),
                            DistanceTypes.L2, 0, .01, .01)
                    );
                }
                else
                {
                    var p = points.Take(index2 + 1)
                        .Concat(points
                            .Skip(index1 - index2 - 1)
                            .Take(points.Count - index1)
                        ).ToList();

                    approximatedLines.Add(
                        Cv2.FitLine(p, DistanceTypes.L2, 0, .01, .01)
                    );
                }
            }
            return approximatedLines;
        }

        private List<Point2d> CalculateIntersections(List<Line2D> lines)
        {
            var points = new List<Point2d>();
            for (int i = 0; i < lines.Count; i++)
            {
                points.Add(
                    Helpers.CalculateIntersection(lines[i], lines[(i + 1) % lines.Count])
                );
            }
            return points;
        }

        private void Draw(List<Line2D> lines, List<Point2d> topPoints, Point2d center, List<Point> polygonPoints)
        {
            var img = Image.Clone();
            const int coef = 1000;

            foreach (var line in lines)
            {
                Cv2.Line(img, new Point(line.X1 + coef * line.Vx, line.Y1 + coef * line.Vy),
                    new Point(line.X1 - coef * line.Vx, line.Y1 - coef * line.Vy), new Scalar(255, 100, 10));
            }
            foreach (var point in polygonPoints)
            {
                Cv2.Circle(img, point, 3, new Scalar(10, 200, 50));
            }
            foreach (var point in topPoints)
            {
                Cv2.Circle(img, new Point(point.X, point.Y), 7, new Scalar(10, 10, 200));
            }
            Cv2.Circle(img, new Point(center.X, center.Y), 7, new Scalar(200, 10, 200));
            Cv2.ImShow("PolygonPoints " + Localization, img);
        }
    }
}