﻿using System;
using System.Collections.Generic;
using System.Linq;
using ImageProcessing.DistributedAlgorithms.MarkerGenerating;
using OpenCvSharp;
using YSoft.Rqa.VisualProwess.AutoCalibration.DetectionRefine;
using YSoft.Rqa.VisualProwess.AutoCalibration.Exceptions;

namespace YSoft.Rqa.VisualProwess.AutoCalibration.MarkerDetecting
{
    internal class MarkerDetector
    {
        internal MarkerTemplate MarkerTemplate { get; set; }
        internal Mat Image { get; set; }
        internal Mat PreprocessedImage {get; set; }
        internal int ThresholdCoefIndex { get; set; }
        internal double Threshold { get; set; }
        internal Marker Marker { get; private set; }

        private const double MaxRelativeDistanceOfTopPoints = 1 / 2.0;
        private const double MinRelativeDistanceOfTopPoints = 1 / 8.0;
        private readonly double[] _thresholdCoefs = { 1, 2, .3, .5, 3.5, .12, 6 };

        internal MarkerDetector(MarkerTemplate markerTemp, Mat image)
        {
            MarkerTemplate = markerTemp;
            Image = image.Clone();
            ThresholdCoefIndex = 0;

            Mat retypedImage = new Mat(image.Size(), MatType.CV_8UC3);
            Image.ConvertTo(retypedImage, MatType.CV_8UC3);
            PreprocessedImage = retypedImage
                .CvtColor(ColorConversionCodes.RGB2GRAY, 1)
                //.GaussianBlur(Helpers.GetGaussFilterSize(), Helpers.Sigma, Helpers.Sigma)
                ;
            Threshold = Cv2.Threshold(PreprocessedImage, new Mat(), 0, 255,
                ThresholdTypes.Otsu | ThresholdTypes.Binary);
        }

        internal Marker Detect(bool improveDetection = false)
        {
            if(Image == null)
                throw new ArgumentNullException("Mat image cannot be null!");
            
            while(true)
            {
                try
                {
                    Mat thresholdedImage =
                        PreprocessedImage.Threshold(Threshold * _thresholdCoefs[ThresholdCoefIndex], 255,
                            ThresholdTypes.Binary);
                    //ShowPreprocessImage(thresholdedImage);
                    var allContours = FindAllContours(thresholdedImage);
                    if (allContours.Count < 4)
                        throw new MarkerCannotBeDetectedException("Marker cannot be detected. Couldn't find enough contours.");
                    FindMarker(allContours);
                    break;
                }
                catch (MarkerCannotBeDetectedException)
                {
                    if (++ThresholdCoefIndex >= _thresholdCoefs.Length)
                        throw;
                }
            }
            if (improveDetection)
            {
                InflectionPointCorrection();
            }

            return Marker;
        }

        private List<List<Point>> FindAllContours(Mat image)
        {
            image.FindContours(out var allContours, out var _, RetrievalModes.List, ContourApproximationModes.ApproxNone);
            var allContoursInList = allContours.Select(line => line.ToList()).ToList();
            return allContoursInList;
        }

        private void FindMarker(List<List<Point>> contours)
        {
            var prefilteredContours = PrefilterContours(contours);
            if (prefilteredContours.Count < 4)
                throw new MarkerCannotBeDetectedException("Marker cannot be detected. There isn't enough contours after prefiltering.");
            var contoursForEval = prefilteredContours.Select(c => new ContourForEvaluation(c)).ToList();
            contoursForEval.ForEach(c => c.Init());
            var postfilteredContours = PostfilterContours(contoursForEval);
            var cornerShapes = MarkerTemplate.CornerShapes.Count > 0 ? FindCornerShapes(postfilteredContours) : new List<CornerShape>();
            
            Marker = new Marker(cornerShapes, MarkerTemplate);
        }

        private List<ContourForEvaluation> PostfilterContours(List<ContourForEvaluation> contours)
        {
            var res = new List<ContourForEvaluation>();
            foreach (var contour in contours)
            {
                if (contour.Rmse.Contains(double.NaN))
                    continue;
                res.Add(contour);
            }
            return res;
        }

        private List<List<Point>> PrefilterContours(List<List<Point>> contours)
        {
            var rightSizeContours = TakeContoursBiggerThan(TakeContoursSmallerThan(contours));
            var filteredContours = new List<List<Point>>();
            foreach (var contour in rightSizeContours)
            {
                var topPointIndexes = FindTopPointIndexes(out var rotatedContour, contour);
                if (!AreTopPointsRegularlyDistributed(rotatedContour, topPointIndexes))
                    continue;
                filteredContours.Add(rotatedContour);
            }
            return filteredContours;
        }

        private List<List<Point>> TakeContoursBiggerThan(List<List<Point>> contours)
        {
            return contours
                .Where(contour => contour.Count > Helpers.GetMinimalPerimeterOfCornerContour(Image.Size()))
                .ToList();
        }

        private List<List<Point>> TakeContoursSmallerThan(List<List<Point>> contours)
        {
            return contours
                .Where(contour => contour.Count < Helpers.GetMaximalPerimeterOfCornerContour(Image.Size()))
                .ToList();
        }

        private int[] FindTopPointIndexes(out List<Point> rotatedContour, List<Point> contour)
        {
            //[right,down,left,up]
            int[] indexes = new int[4];
            var offset = contour.IndexOf(contour.OrderByDescending(p => p.X).First());
            rotatedContour = contour.RotateList(offset);

            indexes[0] = 0;
            indexes[1] = rotatedContour.IndexOf(rotatedContour.OrderByDescending(p => p.Y).First());
            indexes[2] = rotatedContour.IndexOf(rotatedContour.OrderBy(p => p.X).First());
            indexes[3] = rotatedContour.IndexOf(rotatedContour.OrderBy(p => p.Y).First());

            return indexes;
        }

        private bool AreTopPointsRegularlyDistributed(List<Point> contour, int[] topPointIndexes)
        {
            for (int i = 0; i < topPointIndexes.Length - 1; i++)
            {
                if ((topPointIndexes[i + 1] - topPointIndexes[i]) / (double)contour.Count >
                    MaxRelativeDistanceOfTopPoints
                    ||
                    (topPointIndexes[i + 1] - topPointIndexes[i]) / (double)contour.Count <
                    MinRelativeDistanceOfTopPoints
                )
                {
                    return false;
                }
            }
            return true;
        }

        private List<CornerShape> FindCornerShapes(List<ContourForEvaluation> contours)
        {
            var cornerShapes = new List<CornerShape>(MarkerTemplate.CornerShapes.Count);
            
            foreach (CornerShapeTemplate corner in this.MarkerTemplate.CornerShapes)
            {
                var cornerContour = contours
                    .Where(c => c.TopPoints.Count == 5)
                    .Where(c => (int) c.GetLocalizationOfCenterPoint() ==
                                Helpers.GetOppositeCorner(corner.ScreenLocalization))
                    .OrderBy(c => c.Rmse.Sum())
                    .FirstOrDefault();

                if (cornerContour == null)
                    throw new MarkerCannotBeDetectedException($"Marker cannot be detected. Couldn't detect shape in {corner.ScreenLocalization} corner");

                cornerShapes.Add(new CornerShape(cornerContour,corner.ScreenLocalization,this.Image));
                cornerShapes.Last().Init();
            }
            //DrawAllContoursForEval(cornerShapes.Select(c => c.Contour).ToList());
            return cornerShapes;
        }

        internal void DrawAllContours(List<List<Point>> contours)
        {    
            for (int i = 0; i <contours.Count; i++)
            {
                var img = Image.Clone();
                Cv2.DrawContours(img, contours, i, new Scalar(0, 0, 255));
                Cv2.ImShow("contours"+i, img);
            }
        }

        internal void DrawContour(List<Point> contour)
        {
            var img = Image.Clone();
            Cv2.DrawContours(img, new List<List<Point>> {contour}, 0, new Scalar(0, 0, 255));
            Cv2.ImShow("contour", img);
        }

        internal void DrawAllContoursForEval(List<ContourForEvaluation> contours)
        {
            var i = 0;
            foreach (var contour in contours)
            {
                var img = Image.Clone();
                Cv2.DrawContours(img, new List<List<Point>> { contour.AllPoints }, 0, new Scalar(0, 0, 255));
                Cv2.ImShow("contour"+i++, img);
            }
        }

        internal void DrawContourForEval(ContourForEvaluation contour)
        {
            var img = Image.Clone();
            Cv2.DrawContours(img, new List<List<Point>> { contour.AllPoints }, 0, new Scalar(0, 0, 255));
            Cv2.ImShow("contour", img);
        }

        internal void ShowPreprocessImage(Mat thresholdedImage)
        {
            var img = Image.Clone();
            for (int i = 0; i < img.Rows; i++)
            {
                for (int j = 0; j < img.Cols; j++)
                {
                    if ((byte)thresholdedImage.At<int>(i, j) == 0)
                    {
                        Vec3b color = img.Get<Vec3b>(i, j);
                        color.Item0 = 0;
                        color.Item1 = 0;
                        color.Item2 = 255;
                        img.Set<Vec3b>(i, j, color);
                    }
                }
            }
            Cv2.ImShow("mergedImgs", img);
            Cv2.ImWrite("mergedImg.bmp", img);
        }

        private void InflectionPointCorrection()
        {
            foreach (var shape in Marker.CornerShapes)
            {
                var corrector = new DetectionRefiner(shape, Image);
                corrector.RefineShape();
            }
        }
    }
}