﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Threading.Tasks;
using ImageProcessing.DistributedAlgorithms;
using ImageProcessing.DistributedAlgorithms.MarkerGenerating;
using OpenCvSharp;
using PdfSharp;
using PdfSharp.Drawing;
using PdfSharp.Pdf;
using YSoft.Rqa.VisualProwess.AutoCalibration.Db;
using YSoft.Rqa.VisualProwess.AutoCalibration.Db.Entities;
using Point = OpenCvSharp.Point;


namespace YSoft.Rqa.VisualProwess.AutoCalibration.RobotCalibration
{
    public class ChessboardGenerator
    {
        private readonly CalibrationContext _calibrationContext;

        private ChessboardModel _model;

        public ChessboardGenerator(CalibrationContext calibrationContext)
        {
            _calibrationContext = calibrationContext;
        }

        ///Method GeneratePdf() generates .pdf file
        ///with image of chessboard for Robot Calibration reffering to given Robot. File is stored
        ///in local Temp folder.
        /// <returns>Method returns FileInfo object of generated file.</returns>
        ///
        ///Arguments:
        ///  <param name="rbtId">ID of used Robot</param>
        public async Task<FileInfo> GenerateChessboardPdf(int rbtId)
        {
            _model = (await Robot.GetById(rbtId, _calibrationContext)).ChessboardModel;

            var tempPath = Path.Combine(Path.GetTempPath(), "AutoCalibration");
            Directory.CreateDirectory(tempPath);

            var randomFileName = Path.GetFileNameWithoutExtension(Path.GetRandomFileName());
            var imgPath = Path.Combine(tempPath, randomFileName + ".bmp");
            var pdfPath = Path.Combine(tempPath, randomFileName + ".pdf");
            var frame = DrawChessboard();
            Cv2.ImWrite(imgPath, frame);

            PdfDocument pdf = new PdfDocument();
            pdf.Info.Title = "Chessboard";
            PdfPage page = pdf.AddPage();
            page.Size = PageSize.A4;
            page.Orientation = PageOrientation.Portrait;

            XGraphics gfx = XGraphics.FromPdfPage(page);
            XImage xImg = XImage.FromFile(imgPath);
            var offset = 50;
            gfx.DrawImage(xImg, offset, offset, XUnit.FromMillimeter(_model.GetSizeInMm().Width), XUnit.FromMillimeter(_model.GetSizeInMm().Height));
            pdf.Save(pdfPath);

            return new FileInfo(pdfPath);
        }

        private Mat DrawChessboard()
        {
            var pxlSquareSize = _model.SquareSize * Helpers.Mm2PxlDrawCoef;
            var size = new Size(pxlSquareSize * (_model.NumberOfInnerPoints.Width+1),
                pxlSquareSize * (_model.NumberOfInnerPoints.Height+1));
            var img = new Mat(size, MatType.CV_8UC3, Scalar.White);

            var squares = new List<List<Point>>();
            for (int i = 0; i <= _model.NumberOfInnerPoints.Width; i++)
            {
                for (int j = 0; j <= _model.NumberOfInnerPoints.Height; j++)
                {
                    if ((i + j) % 2 != 0) continue;
                    var square = new List<Point>
                    {
                        new Point(i * pxlSquareSize, j * pxlSquareSize),
                        new Point(i * pxlSquareSize, (j + 1) * pxlSquareSize -1),
                        new Point((i + 1) * pxlSquareSize -1, (j + 1) * pxlSquareSize -1),
                        new Point((i + 1) * pxlSquareSize -1, j * pxlSquareSize)
                    };
                    squares.Add(square);
                }
            }
            Cv2.FillPoly(img,squares,Scalar.Black);
            return img;
        }

        internal async Task<FileInfo> GenerateMarkerChessboardTest(int rbtId, int devId, Size screenSize, Point distance)
        {
            var device = await Device.GetById(devId, _calibrationContext);
            var chesboardInfo = await GenerateChessboardPdf(rbtId);
            var chessPath = Path.Combine(chesboardInfo.DirectoryName,
                Path.GetFileNameWithoutExtension(chesboardInfo.Name) + ".bmp");
            var chessImg = Cv2.ImRead(chessPath);

            var markerGenerator = new MarkerGenerator();
            string markerPath;
            using (var markerStream = markerGenerator.GeneratePdf(device.MinDistanceAroundScreen, screenSize))
            {
                markerPath = Path.Combine(Path.GetDirectoryName(markerStream.Name),
                    Path.GetFileNameWithoutExtension(markerStream.Name) + ".bmp");
            }
            var markerImg = Cv2.ImRead(markerPath);
            var marker = device.GetMarkerTemplate;

            var screenSizeInPxl = screenSize.Multiply(Helpers.Mm2PxlDrawCoef);
            var screen2MarkerDiff = (markerImg.Width - screenSizeInPxl.Width);
            var distPxl = distance * Helpers.Mm2PxlDrawCoef;

            var img = markerImg.Clone();
            var chessRect = new Rect(
                new Point(screen2MarkerDiff / 2 + distPxl.X,
                    img.Size().Height - screen2MarkerDiff / 2 - distPxl.Y - chessImg.Size().Height),
                chessImg.Size());
            chessImg.CopyTo(img.SubMat(chessRect));
            Cv2.ImShow("olaolaola", img);

            var tempPath = Path.Combine(Path.GetTempPath(), "AutoCalibration");
            Directory.CreateDirectory(tempPath);

            var randomFileName = Path.GetFileNameWithoutExtension(Path.GetRandomFileName());
            var imgPath = Path.Combine(tempPath, randomFileName + ".bmp");
            var pdfPath = Path.Combine(tempPath, randomFileName + ".pdf");
            Cv2.ImWrite(imgPath, img);

            PdfDocument pdf = new PdfDocument();
            pdf.Info.Title = "Test";
            PdfPage page = pdf.AddPage();
            page.Size = PageSize.A4;
            page.Orientation = PageOrientation.Landscape;

            XGraphics gfx = XGraphics.FromPdfPage(page);
            XImage xImg = XImage.FromFile(imgPath);
            var offset = 50;
            var frameSizeInMm = marker.GetFrameSizeInMm(screenSize);
            gfx.DrawImage(xImg, offset, offset, XUnit.FromMillimeter(frameSizeInMm.Width),
                XUnit.FromMillimeter(frameSizeInMm.Height));
            pdf.Save(pdfPath);

            return new FileInfo(pdfPath);
        }

        internal async Task<FileInfo> GenerateChessboardsTest(int rbtId, Point distance, bool rotate90)
        {
            _model = (await Robot.GetById(rbtId, _calibrationContext)).ChessboardModel;

            var tempPath = Path.Combine(Path.GetTempPath(), "AutoCalibration");
            Directory.CreateDirectory(tempPath);

            var randomFileName = Path.GetFileNameWithoutExtension(Path.GetRandomFileName());
            var imgPath = Path.Combine(tempPath, randomFileName + ".bmp");
            var imgRotPath = Path.Combine(tempPath, randomFileName + "Rot.bmp");
            var pdfPath = Path.Combine(tempPath, randomFileName + ".pdf");
            var frame = DrawChessboard();
            var rotFrame = rotate90 ? frame.Rotate90() : frame.Clone();
            Cv2.ImWrite(imgPath, frame);
            Cv2.ImWrite(imgRotPath, rotFrame);
            
            PdfDocument pdf = new PdfDocument();
            pdf.Info.Title = "ChessboardTest";
            PdfPage page = pdf.AddPage();
            page.Size = PageSize.A4;
            page.Orientation = PageOrientation.Landscape;

            XGraphics gfx = XGraphics.FromPdfPage(page);
            XImage xImg = XImage.FromFile(imgPath);
            XImage xRotImg = XImage.FromFile(imgRotPath);
            var offset = 50;
            gfx.DrawImage(xImg, offset, offset, XUnit.FromMillimeter(_model.GetSizeInMm().Width), XUnit.FromMillimeter(_model.GetSizeInMm().Height));
            var toX = rotate90
                ? XUnit.FromMillimeter(_model.GetSizeInMm().Height)
                : XUnit.FromMillimeter(_model.GetSizeInMm().Width);
            var toY = rotate90
                ? XUnit.FromMillimeter(_model.GetSizeInMm().Width)
                : XUnit.FromMillimeter(_model.GetSizeInMm().Height);
            gfx.DrawImage(xRotImg, offset+ XUnit.FromMillimeter(distance.X), offset+ XUnit.FromMillimeter(distance.Y), toX, toY);
            pdf.Save(pdfPath);

            return new FileInfo(pdfPath);
        }
    }
}
