﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Threading.Tasks;
using FSLibrary;
using ImageProcessing.DistributedAlgorithms.MarkerGenerating;
using MathNet.Numerics.LinearAlgebra;
using OpenCvSharp;
using YSoft.Rqa.VisualProwess.AutoCalibration.Db;
using YSoft.Rqa.VisualProwess.AutoCalibration.Db.Entities;
using YSoft.Rqa.VisualProwess.AutoCalibration.Exceptions;
using YSoft.Rqa.VisualProwess.AutoCalibration.MarkerDetecting;

namespace YSoft.Rqa.VisualProwess.AutoCalibration.RobotCalibration
{
    public class RobotCalibrator: IDisposable
    {
        //Class has 1 public method Calibrate() which performs Robot arm calibration.
        //First the marker on the Device screen is detected, then the chessboard on Robot arm
        //is detected, finally the transform matrix is calculated by GaussNewton iteration method
        //and the result is returned.
        //Method Calibrate() is not time determinant, meaning everytime it takes different amount
        //of time to calculate the result. Nevertheless there is a limit for max number of iterations
        //so it is stable.
        //
        //Initialization must contain:
        //  ID of Robot, Device and Camera used,
        //  Size of Device screen in format (width,height) in mm,
        //  Image of class Mat of MatType.CV_8UC3 preferably, but should manage all C3 MatTypes.
        //
        //Exceptions:
        //  ArgumentNullException if Image is Null
        //  MarkerCannotBeDetectedException
        //  <not implemented yet> ChessboardCannotBeDetectedException

        internal Device Device { get; set; }
        internal Size2f ScreenSize { get; set; }
        internal Robot Robot { get; set; }
        internal Camera Camera { get; set; }
        internal Mat Image { get; set; }
        private Marker Marker { get; set; }

        private readonly double[] _initialEstimation = { 0, 0, 800.0, 0, 0, 0 };
        private readonly double[] _initialEstimationRot = { 0, 0, 800.0, 0, 0, 1.57 };
        private const int MaxNumberOfIterations = 1000;
        private const double NormCeil = 1e-7;

        private readonly CalibrationContext _calibrationContext;

        public RobotCalibrator(CalibrationContext calibrationContext)
        {
            _calibrationContext = calibrationContext;
        }

        internal async Task Init(int devId, Size screenSize, int robotId, int camId, Mat img)
        {
            Device = await Device.GetById(devId, _calibrationContext);
            ScreenSize = new Size2f(screenSize.Width, screenSize.Height); ;
            Robot = await Robot.GetById(robotId, _calibrationContext);
            Camera = await Camera.GetById(camId, _calibrationContext);
            Image = img.Clone();
        }

        public async Task<Matrix<double>> Calibrate(int devId, Size screenSize, int robotId, int camId, Mat img)
        {
            await Init(devId, screenSize, robotId, camId, img);
            //TODO vyres GetOptimalCameraMatrix if needed (strcit to dovnitr kamery jako novou prop)
            //Camera.IntrinsicMatrix = Cv2.GetOptimalNewCameraMatrix(Camera.GetIntrinsicMatrix(),
            //    Camera.GetDistortionCoefficients(), Image.Size(), 0, Image.Size(), out _);

            var transformParamsPrinterToCamera = ScreenToCamera();
            var transformParamsRobotArmToCamera = RobotToCamera();
            //var transformParamsRobotArmToCamera = new double[]{0,0,0,0,0,0};

            //PrintAllPointsAndLines(transformParamsPrinterToCamera,transformParamsRobotArmToCamera);

            var mat = matrixTransform.transform_rbt_to_printer(transformParamsPrinterToCamera,
                transformParamsRobotArmToCamera);
            return mat;
        }

        internal double[] ScreenToCamera(bool refineDetection = true)
        {
            var markerTemplate = Device.GetMarkerTemplate;
            Marker = new MarkerDetector(markerTemplate, Image).Detect(refineDetection);
            var planePoints = ExtractAllPlanePoints(Marker);
            var markerModel = markerTemplate.GetMarkerModel(ScreenSize.ToSize());
            return GaussNewtonFs(markerModel, planePoints, _initialEstimation, Camera.GetIntrinsicMatrix());
        }

        internal double[,] ExtractAllPlanePoints(Marker marker)
        {
            return marker.CornerShapes.SelectMany(s => 
                s.TopPoints.Concat(new List<Point2d>{s.CenterPoint})
            ).ToList()
            .Select(s => new Point2d(s.X, Image.Height-1 - s.Y))
            .ToList()
            .To2dArray();
        }

        internal double[] RobotToCamera(bool rotateChessPoints = false)
        {
            var isFound = Cv2.FindChessboardCorners(Image, Robot.ChessboardModel.NumberOfInnerPoints.ToOpenCvSize(), out var corners);
            if (!isFound)
                throw new RobotCannotBeDetectedException("Cannot detect all chessboard corners.");

            var subCorners = Cv2.CornerSubPix(Image.CvtColor(ColorConversionCodes.RGB2GRAY, 1), corners,
                new Size(11, 11), new Size(-1, -1),
                new TermCriteria(CriteriaType.Eps | CriteriaType.MaxIter, 10000, 1e-6))
                .ToList()
                .Select(s=> new Point2d(s.X, Image.Height-1 - s.Y))
                .ToArray();

            var initEstimation = rotateChessPoints ? _initialEstimationRot : _initialEstimation;
            //if (rotateChessPoints)
            //{
            //    subCorners = RotatePointsForChessboardsTest(subCorners.ToList()).ToArray();
            //}

            var i = Image.Clone();
            var windowName = Path.GetFileNameWithoutExtension(Path.GetRandomFileName());
            foreach (Point2d point in subCorners)
            {
                var p = new Point(point.X, Image.Height - 1 - point.Y);
                Cv2.Circle(i, p, 3, new Scalar(10, 200, 50));
            }
            
            var robotToCam = GaussNewtonFs(Robot.GetChessboardPoints().To2dArray(), subCorners.To2dArray(), initEstimation, Camera.GetIntrinsicMatrix());
            //for (int i = 0; i < 3; i++)
            //{
            //    robotToCam[i] = robotToCam[i] + Robot.TranslationChessboardToArm[i];
            //}

            //var backProjectedPoints =
            //    BackProjection(robotToCam, Camera.GetIntrinsicMatrix(), Robot.GetChessboardPoints().To2dArray());
            //for (int j = 0; j < backProjectedPoints.GetLength(1); j++)
            //{
            //    var point = new Point2d(backProjectedPoints[0, j], Image.Height - 1 - backProjectedPoints[1, j]);
            //    Cv2.Circle(i, (Point)point, 2, new Scalar(200, 50, 10));
            //}
            //Cv2.ImShow(windowName, i);
            return robotToCam;
        }

        public static double[] GaussNewtonFs(double[,] x3D, double[,] x2D, double[] initEstimation, double[,] K)
        {
            double norm = 100.0;
            var params_i = initEstimation; //tx,ty,tz,alpha,beta,gamma
            var params_i1 = new double[6];
            int i = 0;

            while (norm > NormCeil && i < MaxNumberOfIterations)
            {
                i++;
                var cond = i % 1 == 0;
                params_i1 = GaussNewton.one_iter(params_i, K, x3D, x2D, cond);
                norm = Helpers.NormL2(params_i, params_i1);
                params_i = params_i1;

                if(norm <= NormCeil)
                {
                    Console.WriteLine(i); Console.WriteLine("");
                    foreach (var d in params_i)
                    {
                        Console.WriteLine(d);
                    }
                    Console.WriteLine(norm + " - L2 distance of i-th and (i+1)th  estimation");
                    var xx = Helpers.NormL2(params_i, new[] {0.0, 0.0, 0.0, 0.0, 0.0, 0.0});
                    Console.WriteLine(norm / xx + " - normed L2 distance of i-th and (i+1)th  estimation");
                    Console.WriteLine("___________________");
                }

                //if (norm <= Helpers.NormCeil)
                //{
                //    foreach (var d in params_i)
                //    {
                //        Console.WriteLine(d);
                //    }
                //}
            }
            return params_i1;
        }

        private void PrintAllPointsAndLines(double[] printerParams, double[] robotParams)
        {
            // red points = backprojected
            // green points = detected

            var i = Image.Clone();
            int lineKoef = 500;
            var lines = Marker.CornerShapes.SelectMany(s => s.Contour.Lines);
            foreach (var line in lines)
            {
                Cv2.Line(i, new Point(line.X1 + lineKoef * line.Vx, line.Y1 + lineKoef * line.Vy),
                    new Point(line.X1 - lineKoef * line.Vx, line.Y1 - lineKoef * line.Vy), new Scalar(255, 100, 10));
            }
            
            var markerPoints = Marker.CornerShapes.SelectMany(s =>
                s.TopPoints.Concat(new List<Point2d> {s.CenterPoint})
            ).ToList();
            Cv2.FindChessboardCorners(Image, Robot.ChessboardModel.NumberOfInnerPoints.ToOpenCvSize(), out var chessPoints);
            var subCorners = Cv2.CornerSubPix(Image.CvtColor(ColorConversionCodes.RGB2GRAY, 1), chessPoints,
                    new Size(11, 11), new Size(-1, -1),
                    new TermCriteria(CriteriaType.Eps | CriteriaType.MaxIter, 10000, 1e-6))
                .ToList();
            var points = markerPoints.Concat(subCorners.ToArray().ConvertToPoint2d().ToList()).ToList();
            foreach (Point point in points)
            {
                Cv2.Circle(i, point, 3, new Scalar(10, 200, 50));
            }

            var backprojectedPoints = GetBackProjectedPoints(printerParams, robotParams);
            
            for (int j = 0; j < backprojectedPoints.Item1.GetLength(1); j++)
            { 
                var point = new Point2d(backprojectedPoints.Item1[0, j], Image.Height - 1 - backprojectedPoints.Item1[1, j]);
                Cv2.Circle(i, (Point)point, 2, new Scalar(10, 50, 200));
            }
            for (int j = 0; j < backprojectedPoints.Item2.GetLength(1); j++)
            {
                var point = new Point2d(backprojectedPoints.Item2[0, j], Image.Height - 1 - backprojectedPoints.Item2[1, j]);
                Cv2.Circle(i, (Point)point, 2, new Scalar(10, 50, 200));
            }

            Cv2.ImShow("Points", i);
            Cv2.ImWrite("points.bmp", i);
        }

        private Tuple<double[,], double[,]> GetBackProjectedPoints(double[] printerParams, double[] robotParams)
        {
            var robotPoints =
                BackProjection(robotParams, Camera.GetIntrinsicMatrix(), Robot.GetChessboardPoints().To2dArray());
            var markerPoints = BackProjection(printerParams, Camera.GetIntrinsicMatrix(),
                Device.GetMarkerTemplate.GetMarkerModel(ScreenSize.ToSize()));
            return Tuple.Create(robotPoints, markerPoints);
        }

        internal static double[,] BackProjection(double[] prms, double[,] k, double[,] x3D)
        {
            return matrixTransform.backprojection(prms, k, x3D);
        }

        internal double[] SolveChessByPnP(bool rotateChessPoints = false)
        {
            var objectPoints = Robot.GetChessboardPoints()
                .Select(p => new Point3f((float) p.Val0, (float) p.Val1, (float) p.Val2))
                .ToList();

            var isFound = Cv2.FindChessboardCorners(Image, Robot.ChessboardModel.NumberOfInnerPoints.ToOpenCvSize(), out var corners);
            if (!isFound)
                throw new RobotCannotBeDetectedException("Cannot detect all chessboard corners.");

            var subCorners = Cv2.CornerSubPix(Image.CvtColor(ColorConversionCodes.RGB2GRAY, 1), corners,
                    new Size(11, 11), new Size(-1, -1),
                    new TermCriteria(CriteriaType.Eps | CriteriaType.MaxIter, 10000, 1e-6))
                .ToList()
                .Select(s => new Point2f(s.X, Image.Height - 1 - s.Y))
                .ToList();

            if (rotateChessPoints)
                subCorners = RotatePointsForChessboardsTest(subCorners);

            var tvec = new double[]{_initialEstimation[0], _initialEstimation[1], _initialEstimation[2]};
            var rvec = new double[]{_initialEstimation[3], _initialEstimation[4], _initialEstimation[5]};

            Cv2.SolvePnP(objectPoints,subCorners,Camera.GetIntrinsicMatrix(),Camera.GetDistortionCoefficients().ToList(),out rvec, out tvec, false);
            return tvec.Concat(rvec).ToArray();
        }

        private List<Point2d> RotatePointsForChessboardsTest(List<Point2d> input)
        {
            var output = new List<Point2d>();
            for (int i = 0; i < 5; i++)
            {
                for (int j = 3; j >= 0; j--)
                {
                    output.Add(input[5*j+i]);
                }
            }
            return output;
        }

        private List<Point2f> RotatePointsForChessboardsTest(List<Point2f> input)
        {
            return RotatePointsForChessboardsTest(input.Select(x => new Point2d(x.X,x.Y)).ToList())
                .Select(x => new Point2f((float)x.X,(float)x.Y))
                .ToList();
        }

        public void Dispose()
        {
            _calibrationContext.Dispose();
        }
    }
}
